
#ifndef __SMLL_CPP__
#define __SMLL_CPP__

#ifdef __SMLLDEF__

	#include "link.h"
	#include <stdio.h>
	#include "smll.h"
	#include <new.h>

	static CRITICAL_SECTION	_smll_cs0;
	static char temp_file[256];
	static unsigned long temp_line;

	static LinkedList<_NEWSTRUCT*> _dbg_ml;
	static char *typestr[] = {"new", "new[]", "malloc()", "calloc()", "realloc()",
		"delete", "delete[]", "free()" };

	#define ENTERCS()	EnterCriticalSection(&_smll_cs0)
	#define LEAVECS()	LeaveCriticalSection(&_smll_cs0)

	void SMLLInit(int autoreport)
	{
		InitializeCriticalSection(&_smll_cs0);

		if (autoreport)
			atexit(SMLLReport);
	}

	void SMLLReport()
	{
		DeleteCriticalSection(&_smll_cs0);
		smll_report();
	}

	void* operator new(size_t size, const char *file, unsigned long line)
	{
		ENTERCS();
		void *ptr = malloc(size);
		smll_alloc_info(ptr, size, file, line, 0, 5);
		LEAVECS();
		return ptr;
	}

	void* operator new[](size_t size, const char *file, unsigned long line)
	{
		ENTERCS();
		void *ptr = malloc(size);
		smll_alloc_info(ptr, size, file, line, 1, 6);
		LEAVECS();
		return ptr;
	}

	void operator delete(void *p)
	{
		smll_dealloc_info(p, temp_file, temp_line, 5);
		free(p);
		LEAVECS();
	}

	void operator delete[](void *p)
	{
		smll_dealloc_info(p, temp_file, temp_line, 6);
		free(p);
		LEAVECS();
	}

	void smll_alloc_info(void *ptr, size_t size, const char *file, unsigned long line, int ntype, int dtype)
	{
		static _NEWSTRUCT *ns;

		ns = new _NEWSTRUCT;
		ns->p = (unsigned char*)ptr;
		ns->line = line;
		ns->size = size;
		ns->ntype = ntype;
		ns->dtype = dtype;
		ns->fn = (char*)malloc(strlen(file)+1);
		strcpy(ns->fn, file);
		_dbg_ml.Add(ns);
	}

	void* malloc(size_t size, const char *file, unsigned long line)
	{
		ENTERCS();
		void *ptr = malloc(size);
		smll_alloc_info(ptr, size, file, line, 2, 7);
		LEAVECS();
		return ptr;
	}

	void* calloc(size_t num, size_t size, const char *file, unsigned long line)
	{
		ENTERCS();
		void *ptr = calloc(num, size);
		smll_alloc_info(ptr, num*size, file, line, 3, 7);
		LEAVECS();
		return ptr;
	}

	void* realloc(void *mem, size_t size, const char *file, unsigned long line)
	{
		ENTERCS();
		void *ptr = realloc(mem, size);
		if (size == 0)
			smll_dealloc_info(mem, file, line, 7); // the block is freed.
		else
		{
			smll_dealloc_info(mem, file, line, 7);
			smll_alloc_info(ptr, size, file, line, 4, 7);
		}
		LEAVECS();
		return ptr;
	}

	void smll_prepare_dealloc(const char *file, unsigned long line)
	{
		ENTERCS();
		strcpy(temp_file, file);
		temp_line = line;
	}

	void smll_dealloc_info(void *p, const char *file, unsigned long line, int dtype)
	{
		static _NEWSTRUCT *ref;
		static LinkedList<_NEWSTRUCT*>::LINKEDLIST *chain;
		int found = 0;

		chain = _dbg_ml.FindFirst();
		while (1)
		{
			ref = _dbg_ml.GetData(chain);
			if (ref == NULL)
				break;

			if (ref->p == p)
			{
				if (ref->dtype != dtype)
				{
					char buf[500];
					sprintf(buf, "file: %s, alloc at: %d\nfile: %s, dealloc tried at: %d\n\nInvalid deallocation. It should be '%s'.",
						ref->fn, ref->line, file, line, typestr[ref->dtype]);
					::MessageBox(NULL, buf, "SMLL warning", MB_OK);
					found = 1;	// trick: to go out the function.
					break;
				}

				found = 1;	// found
				free(ref->fn);	// dealloc filename str.
				_dbg_ml.Delete(chain);
				break;
			}

			chain = _dbg_ml.FindNext(chain);
		}

		if (!found)
		{
			char buf[500];
			sprintf(buf, "file: %s, dealloc tried at: %d\n\nIt is not allocated by SMLL or was already freed.",
				file, line);
			::MessageBox(NULL, buf, "SMLL warning", MB_OK);
		}
	}

	void free(void *p, const char *file, unsigned long line)
	{
		ENTERCS();
		smll_dealloc_info(p, file, line, 7);
		free(p);
		LEAVECS();
	}

	void smll_report()
	{
		_NEWSTRUCT *ref;
		int lcount = 0;
		int i, dump_len = 15;
		FILE *fp;

		fp = fopen("_smll.txt", "wt");

		fprintf(fp, "- memory leak states -\n");

		while (ref = _dbg_ml.GetData(_dbg_ml.FindFirst()))
		{
			fprintf(fp, "FILE: %s\n", ref->fn);
			fprintf(fp, "(%d) USE: %s, LINE #: %d, PTR: 0x%08X, %d bytes long.\n", 
				++lcount, typestr[ref->ntype],
				ref->line, ref->p, ref->size);
			if (ref->p == NULL)
				fprintf(fp, "memory allocation was failed!\n");
			else
			{
				#pragma warning( disable : 4018 )
				if (ref->size < dump_len)
					dump_len = ref->size;
				#pragma warning( default : 4018 )
				fprintf(fp, " DUMP: [ ");
				for (i=0; i<dump_len; i++)
					fprintf(fp, "%c", 
						(ref->p[i] < 30) ? ' ' : ref->p[i]);
				fprintf(fp, " ]  ");
				for (i=0; i<dump_len; i++)
					fprintf(fp, "%02X ", ref->p[i]);
				fprintf(fp, "\n");
			}
			free(ref->fn);	// dealloc filename str.
			_dbg_ml.DeleteFirst();
		}

		fclose(fp);
	}

#endif

#endif

